unit log_unit;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, fgl,
  intstring_map;

type

  TLogManager = class;
  TLogClient = class;

  { TLogMessage }

  TLogMessage = object
  public
    procedure Meow(const aText: string; const aKind: integer;
      const aWho: TObject); inline;
  private
    fText: string;
    fKind: integer;
    fWho: TObject;
  public
    property text: string read fText;
    property kind: integer read fKind;
    property who: TObject read fWho;
  end;

  PLogMessage = ^TLogMessage;

  { TLog }

  TLog = class
  public
    class function fatal: integer;
    class function error: integer;
    class function warn: integer;
    class function info: integer;
    class function debug: integer;
    class function trace: integer;

  public
    class function defaultFormat(const aMessage: PLogMessage): string;
    class procedure addDefaultKinds(const aMap: TIntstringMap);
  end;


  { ILogClient }

  ILogClient = interface
    procedure m(const aKind: integer; const aText: string);
    procedure fatal(const aText: string);
    procedure error(const aText: string);
    procedure warn(const aText: string);
    procedure info(const aText: string);
    procedure debug(const aText: string);
    procedure trace(const aText: string);
  end;

  { TLogClient }

  TLogClient = class(TInterfacedObject, ILogClient)
  public
    constructor Create(const aOwner: TObject; const aManager: TLogManager);
  private
    fOwner: TObject;
    fManager: TLogManager;
  public
    property owner: TObject read fOwner;
    property manager: TLogManager read fManager;

    {%REGION sendmessage methods}
    procedure m(const aKind: integer; const aText: string);
    procedure fatal(const aText: string);
    procedure error(const aText: string);
    procedure warn(const aText: string);
    procedure info(const aText: string);
    procedure debug(const aText: string);
    procedure trace(const aText: string);
    {%ENDREGION }

    destructor Destroy; override;
  end;


  { ILogAppender }

  ILogAppender = interface
    ['{4A51DFF3-B36F-4330-8E58-2E9C524BEAC4}']

    function appendMessage(const aMessage: PLogMessage): boolean;
  end;

  TLogAppenders = specialize TFPGList<ILogAppender>;


  { TLogAppender }

  TLogAppender = class(TInterfacedObject)
  public
    class function Meow: ILogAppender;
  end;

  TLogAppenderClass = class of TLogAppender;


  { TLogManager }

  TLogManager = class(TComponent)
  public
    constructor Create(aOwner: TComponent); override;
  private
    procedure startup;
  private
    fAppenders: TLogAppenders;
    fLog: ILogClient;
    procedure clear;
  public
    procedure message(const aMessage: TLogMessage);
    function getLogClient(const aObject: TObject): ILogClient;
    function getLogClient(const aObject: TObject;
      var aLog: ILogClient): ILogClient;
    procedure addAppender(const aClass: TLogAppenderClass);
    procedure addAppender(const anAppender: ILogAppender);
    property appenders: TLogAppenders read fAppenders;
    property log: ILogClient read fLog;
    destructor Destroy; override;
  end;


function defaultLogKinds: TIntstringMap; inline;

implementation

var
  defaultKinds: TIntstringMap = nil;

function defaultLogKinds: TIntstringMap; inline;
begin
  if defaultKinds = nil then
  begin
    defaultKinds:=TIntstringMap.Create(nil);
    TLog.addDefaultKinds(defaultKinds);
  end;
  result:=defaultKinds;
end;


{ TLogMessage }

procedure TLogMessage.Meow(const aText: string; const aKind: integer;
  const aWho: TObject);
begin
  fText:=aText;
  fKind:=aKind;
  fWho:=aWho;
end;


{ TLog }

class function TLog.fatal: integer;
begin
  result:=-100;
end;

class function TLog.error: integer;
begin
  result:=-10;
end;

class function TLog.warn: integer;
begin
  result:=-1;
end;

class function TLog.info: integer;
begin
  result:=0;
end;

class function TLog.debug: integer;
begin
  result:=1;
end;

class function TLog.trace: integer;
begin
  result:=2;
end;

class function TLog.defaultFormat(const aMessage: PLogMessage): string;
var
  kind: string;
begin
  result:='';
  if aMessage = nil then exit('nil log message');

  kind:=defaultLogKinds[ aMessage^.kind ];
  if kind = '' then
    result += 'no_kind'
  else
    result += kind;
  result += ' ';

  if aMessage^.who = nil then
    result += 'Nil object'
  else
    result += aMessage^.who.ClassName;
  result += ': ';

  result += aMessage^.text;
end;

class procedure TLog.addDefaultKinds(const aMap: TIntstringMap);
begin
  if aMap = nil then
    raise Exception.Create('Can not write default log kinds: map is nil');
  aMap[TLog.fatal]:='[FATAL]';
  aMap[TLog.error]:='[X]';
  aMap[TLog.warn]:= '[!]';
  aMap[TLog.info]:= '[*]';
  aMap[TLog.debug]:='[ ]';
  aMap[TLog.trace]:=' . ';
end;


{ TLogClient }

constructor TLogClient.Create(const aOwner: TObject;
  const aManager: TLogManager);
begin
  inherited Create;
  fOwner:=aOwner;
  fManager:=aManager;
end;

procedure TLogClient.m(const aKind: integer; const aText: string);
var
  message: TLogMessage;
begin
  message.Meow(aText, aKind, owner);
  manager.message(message);
end;

procedure TLogClient.fatal(const aText: string);
begin
  m(TLog.fatal, aText);
end;

procedure TLogClient.error(const aText: string);
begin
  m(TLog.error, aText);
end;

procedure TLogClient.warn(const aText: string);
begin
  m(TLog.warn, aText);
end;

procedure TLogClient.info(const aText: string);
begin
  m(TLog.info, aText);
end;

procedure TLogClient.debug(const aText: string);
begin
  m(TLog.debug, aText);
end;

procedure TLogClient.trace(const aText: string);
begin
  m(TLog.trace, aText);
end;

destructor TLogClient.Destroy;
begin
  inherited Destroy;
end;


{ TLogAppender }

class function TLogAppender.Meow: ILogAppender;
var
  o: TLogAppender;
  f: ILogAppender;
  r: boolean;
begin
  o:=self.Create;
  r:=o.GetInterface(ILogAppender, f);
  if not r then
    raise Exception.Create('Unable to meow class: no ILogAppender interface');
  result:=f;
end;


{ TLogManager }

constructor TLogManager.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  clear;
  startup;
end;

procedure TLogManager.startup;
begin
  fAppenders:=TLogAppenders.Create;
  fLog:=getLogClient(self);
end;

procedure TLogManager.clear;
begin
  fAppenders:=nil;
end;

procedure TLogManager.message(const aMessage: TLogMessage);
var
  appender: ILogAppender;
begin
  for appender in appenders do
  begin
    appender.appendMessage(@aMessage);
  end;
end;

function TLogManager.getLogClient(const aObject: TObject): ILogClient;
begin
  result:=TLogClient.Create(aObject, self);
end;

function TLogManager.getLogClient(const aObject: TObject;
  var aLog: ILogClient): ILogClient;
begin
  if not assigned(aLog) then
    aLog:=getLogClient(aObject);
  result:=aLog;
end;

procedure TLogManager.addAppender(const aClass: TLogAppenderClass);
begin
  appenders.add(aClass.Meow);
  log.info('New appender added: ' + aClass.ClassName);
end;

procedure TLogManager.addAppender(const anAppender: ILogAppender);
begin
  appenders.add(anAppender);
  log.info('New appender added');
end;

destructor TLogManager.Destroy;
begin
  appenders.Free;
  inherited Destroy;
end;


finalization
  if defaultKinds <> nil then FreeAndNil(defaultKinds);
end.

