unit uniconfig;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, jsonconf, fgl;

type

  TXConfigKind = (xkInteger, xkString, xkFloat, xkBoolean);
  TXConfigKeyExists = (keYes, keNo, keWrong,
    keNoconfig, keNilitem, keNil, keAgain);

  { TXConfigItem }

  TXConfigItem = class
  public
    constructor Create(aKey: string; aKind: TXConfigKind);
  private
    fKey: string;
    fKind: TXConfigKind;
  public
    property key: string read fKey;
    property kind: TXConfigKind read fKind;
  end;

  TXConfigItems = specialize TFPGList<TXConfigItem>;

  { TXConfig }

  TXConfig = class(TComponent)
  public
    constructor Create(const aOwner: TComponent; const aFilename: string);
      reintroduce;
  private
    procedure startup(const aFilename: string);
  private
    fConf: TJSONConfig;
    fItems: TXConfigItems;
    procedure clear;
    function getStr(const aKey: string): string;
    procedure setStr(const aKey: string; const aValue: string);
    function getInt(const aKey: string): integer;
    procedure setInt(const aKey: string; const aValue: integer);
    function getFlo(const aKey: string): double;
    procedure setFlo(const aKey: string; const aValue: double);
    function getBoo(const aKey: string): boolean;
    procedure setBoo(const aKey: string; const aValue: boolean);
    function keyExists(const aKey: string;
      const aKind: TXConfigKind): TXConfigKeyExists;
    procedure checkKey(const aKey: string; const aKind: TXConfigKind);
    procedure freeItems;
    procedure freeConf;
  protected
    procedure createConfItems; virtual;
  public
    function addItem(const aKey: string;
      const aKind: TXConfigKind): integer;
    procedure resetConfig; virtual;
    property conf: TJSONConfig read fConf;
    property items: TXConfigItems read fItems;
    property str[const aKey: string]: string read getStr write setStr;
    property int[const aKey: string]: integer read getInt write setInt;
    property flo[const aKey: string]: double read getFlo write setFlo;
    property boo[const aKey: string]: boolean read getBoo write setBoo;
    class function KEY_ONCECONFIGURED: string;
    destructor Destroy; override;
  end;

  EXConfig = class(Exception);

implementation

{ TXConfigItem }

constructor TXConfigItem.Create(aKey: string; aKind: TXConfigKind);
begin
  inherited Create;
  fKey:=aKey;
  fKind:=aKind;
end;

{ TXConfig }

constructor TXConfig.Create(const aOwner: TComponent; const aFilename: string);
begin
  inherited Create(aOwner);
  clear;
  startup(aFilename);
end;

procedure TXConfig.startup(const aFilename: string);
begin
  fConf:=TJSONConfig.Create(self);
  if aFilename <> '' then
    conf.Filename:=aFilename;
  fItems:=TXConfigItems.Create;
  createConfItems;
  if not boo[KEY_ONCECONFIGURED] then resetConfig;
end;

procedure TXConfig.clear;
begin
  fConf:=nil;
  fItems:=nil;
end;

function TXConfig.getStr(const aKey: string): string;
begin
  checkKey(aKey, xkString);
  result:=conf.GetValue(aKey, '');
end;

procedure TXConfig.setStr(const aKey: string; const aValue: string);
begin
  checkKey(aKey, xkString);
  conf.setValue(aKey, aValue);
end;

function TXConfig.getInt(const aKey: string): integer;
begin
  checkKey(aKey, xkInteger);
  result:=conf.GetValue(aKey, 0);
end;

procedure TXConfig.setInt(const aKey: string; const aValue: integer);
begin
  checkKey(aKey, xkInteger);
  conf.SetValue(aKey, aValue);
end;

function TXConfig.getFlo(const aKey: string): double;
begin
  checkKey(aKey, xkFloat);
  result:=conf.getValue(aKey, double(0.0));
end;

procedure TXConfig.setFlo(const aKey: string; const aValue: double);
begin
  checkKey(aKey, xkFloat);
  conf.SetValue(aKey, aValue);
end;

function TXConfig.getBoo(const aKey: string): boolean;
begin
  checkKey(aKey, xkBoolean);
  result:=conf.getValue(aKey, false);
end;

procedure TXConfig.setBoo(const aKey: string; const aValue: boolean);
begin
  checkKey(aKey, xkBoolean);
  conf.setValue(aKey, aValue);
end;

function TXConfig.keyExists(const aKey: string;
  const aKind: TXConfigKind): TXConfigKeyExists;
var
  i: TXConfigItem;
begin
  if self = nil then exit(keNil);
  result:=keNo;
  if conf = nil then exit(keNoconfig);
  for i in items do
  begin
    if i = nil then exit(keNilitem);
    if i.key = aKey then
    begin
      if result = keYes then exit(keAgain);
      if i.kind = aKind then
        result:=keYes
      else
        exit(keWrong);
    end;
  end;
end;

procedure TXConfig.checkKey(const aKey: string; const aKind: TXConfigKind);
var
  r: TXConfigKeyExists;
begin
  r:=keyExists(aKey, aKind);
  case r of
    keYes: ;
    keNo: raise EXConfig.Create('No such config key "' + aKey + '"');
    keWrong: raise EXConfig.Create('Wrong key for "' + aKey + '" config item');
    keNoconfig: raise EXConfig.Create('Conf not assigned');
    keNilitem: raise EXConfig.Create('Nil conf item found');
    keNil: raise EXConfig.Create('TXConfig is nil');
    keAgain: raise EXConfig.Create('Duplicate item found');
  end;
end;

procedure TXConfig.freeItems;
var
  i: TXConfigItem;
begin
  for i in items do
    i.Free;
  items.Free;
end;

procedure TXConfig.freeConf;
begin
  conf.Flush;
  conf.Free;
end;

procedure TXConfig.createConfItems;
begin
  addItem(KEY_ONCECONFIGURED, xkBoolean);
end;

function TXConfig.addItem(const aKey: string;
  const aKind: TXConfigKind): integer;
begin
  result:=items.add( TXConfigItem.Create(aKey, aKind) );
end;

procedure TXConfig.resetConfig;
begin
  boo[KEY_ONCECONFIGURED]:=true;
end;

class function TXConfig.KEY_ONCECONFIGURED: string;
begin
  result:='once_configured';
end;

destructor TXConfig.Destroy;
begin
  freeItems;
  freeConf;
  inherited Destroy;
end;

end.

